/* 
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 

#ifndef __INVESTIGATION_DONNEES_CONDITION__
#define __INVESTIGATION_DONNEES_CONDITION__

#include <string>
#include <vector>
#include <src/global/global.h>
namespace Investigation { namespace Donnees { DECL_SHARED_PTR(Condition); } }
#include <src/investigation/donnees/requete.h>
#include <src/investigation/donnees/expressionrationnelle.h>

namespace Investigation
{
	namespace Donnees
	{
		class ConditionEt;
		class ConditionOuInclusif;
		class ConditionOuExclusif;
		class Condition
		{
			friend class ConditionEt;
			friend class ConditionOuInclusif;
			friend class ConditionOuExclusif;
			public:
				Condition(const bool resultat = true) :_resultat(resultat) {};
				virtual ~Condition() {};
				void operator!() { _resultat = not _resultat; };
				virtual bool operator() (const Requete& requete) { return _resultat; };
				virtual ConditionSP simplification() const
				{
					ConditionSP condition(new Condition(*this));
					return condition;
				}
				virtual bool condition_triviale() const { return true; }
				virtual std::pair<bool,std::string> condition_identifiant() const { return std::pair<bool,std::string>(false,""); }
			protected:
				bool _resultat;
			private:
				bool& resultat() { return _resultat; }
		};

		class ConditionIdentifiant : public Condition
		{
			public:
				ConditionIdentifiant(const std::string& identifiant)
				:_identifiant(identifiant) {};
				virtual ~ConditionIdentifiant() {};
				virtual bool operator() (const Requete& requete)
				{
					return _resultat==(requete._occurences.find(_identifiant)!=requete._occurences.end());
				};
				virtual ConditionSP simplification() const
				{
					ConditionSP condition(new ConditionIdentifiant(*this));
					return condition;
				}
				virtual bool condition_triviale() const { return false; }
				virtual std::pair<bool,std::string> condition_identifiant() const { return std::pair<bool,std::string>(true,_identifiant); }
			private:
				const std::string _identifiant;
		};

		template<typename Patron>
			class ConditionExpressionRationnelle : public Condition
		{
			public:
				ConditionExpressionRationnelle(const Patron& patron)
					:_patron(patron) { };
				virtual ~ConditionExpressionRationnelle() { };
				virtual bool operator() (const Requete& requete)
				{
					return _resultat==(_patron(requete));
				};
				virtual ConditionSP simplification() const
				{
					ConditionSP condition(new ConditionExpressionRationnelle<Patron>(*this));
					return condition;
				}
				virtual bool condition_triviale() const { return false; }
			private:
				const Patron _patron;
		};

		class ConditionEt : public Condition
		{
			public:
				ConditionEt(const ConditionSP& gauche, const ConditionSP& droite)
				{
					_operandes.push_back(gauche);
					_operandes.push_back(droite);
				}
				virtual ~ConditionEt() { };
				ConditionEt& operator<<(const ConditionSP& operande)
				{
					_operandes.push_back(operande);
					return *this;
				}
				virtual bool operator() (const Requete& requete)
				{
					for(auto it=_operandes.cbegin() ; it!=_operandes.cend() ; ++it)
					{
						if(not (**it)(requete))
							return not _resultat;
					}
					return _resultat;
				}
				virtual ConditionSP simplification() const
				{
					std::vector<ConditionSP> operandes_simplifies;
					for(auto it=_operandes.cbegin() ; it!=_operandes.cend() ; ++it)
					{
						operandes_simplifies.push_back((*it)->simplification());
					}
					std::vector<ConditionSP> operandes_simplifies_epures;
					for(auto it=operandes_simplifies.cbegin() ; it!=operandes_simplifies.cend() ; ++it)
					{
						if((*it)->condition_triviale())
						{
							if(not (*it)->resultat())
							{
								ConditionSP condition(new Condition(false));
								return condition;
							}
						}
						else
						{
							operandes_simplifies_epures.push_back(*it);
						}
					}
					std::map<std::string, std::pair<std::vector<std::vector<ConditionSP>::const_iterator>, std::vector<std::vector<ConditionSP>::const_iterator> > > identifiants;
					for(auto it=operandes_simplifies_epures.cbegin() ; it!=operandes_simplifies_epures.cend() ; ++it)
					{
						auto identifiant = (*it)->condition_identifiant();
						if(identifiant.first)
						{
							if((*it)->resultat())
							{
								identifiants[identifiant.second].first.push_back(it);
							}
							else
							{
								identifiants[identifiant.second].second.push_back(it);
							}
						}
					}
					for(auto it=identifiants.cbegin() ; it!=identifiants.cend() ; ++it)
					{
						if(not it->second.first.empty() and not it->second.second.empty())
						{
							ConditionSP condition(new Condition(false));
							return condition;
						}
					}
					std::vector<ConditionSP> operandes_simplifies_epures_sans_doublons;
					for(auto it=operandes_simplifies_epures.cbegin() ; it!=operandes_simplifies_epures.cend() ; ++it)
					{
						auto identifiant = (*it)->condition_identifiant();
						if(identifiant.first)
						{
							if(((not identifiants[identifiant.second].first.empty()) and (it==identifiants[identifiant.second].first.front())) or ((not identifiants[identifiant.second].second.empty()) and (it==identifiants[identifiant.second].second.front())))
							{
								operandes_simplifies_epures_sans_doublons.push_back(*it);
							}
						}
						else
						{
							operandes_simplifies_epures_sans_doublons.push_back(*it);
						}
					}
					if(operandes_simplifies_epures_sans_doublons.size()==1)
					{
						return operandes_simplifies_epures_sans_doublons.front();
					}
					ConditionSP condition_simplifiee(new ConditionEt(operandes_simplifies_epures_sans_doublons));
					return condition_simplifiee;
				}
				virtual bool condition_triviale() const { return false; }
			private:
				explicit ConditionEt(const std::vector<ConditionSP>& operandes) :_operandes(operandes) {};
				std::vector<ConditionSP> _operandes;
		};

		class ConditionOuInclusif : public Condition
		{
			public:
				ConditionOuInclusif(const ConditionSP& gauche, const ConditionSP& droite)
				{
					_operandes.push_back(gauche);
					_operandes.push_back(droite);
				}
				virtual ~ConditionOuInclusif() { };
				ConditionOuInclusif& operator<<(const ConditionSP& operande)
				{
					_operandes.push_back(operande);
					return *this;
				}
				virtual bool operator() (const Requete& requete)
				{
					for(auto it=_operandes.cbegin() ; it!=_operandes.cend() ; ++it)
					{
						if((**it)(requete))
							return _resultat;
					}
					return not _resultat;
				}
				virtual ConditionSP simplification() const
				{
					std::vector<ConditionSP> operandes_simplifies;
					for(auto it=_operandes.cbegin() ; it!=_operandes.cend() ; ++it)
					{
						operandes_simplifies.push_back((*it)->simplification());
					}
					std::vector<ConditionSP> operandes_simplifies_epures;
					for(auto it=operandes_simplifies.cbegin() ; it!=operandes_simplifies.cend() ; ++it)
					{
						if((*it)->condition_triviale())
						{
							if((*it)->resultat())
							{
								ConditionSP condition(new Condition(true));
								return condition;
							}
						}
						else
						{
							operandes_simplifies_epures.push_back(*it);
						}
					}
					std::map<std::string, std::pair<std::vector<std::vector<ConditionSP>::const_iterator>, std::vector<std::vector<ConditionSP>::const_iterator> > > identifiants;
					for(auto it=operandes_simplifies_epures.cbegin() ; it!=operandes_simplifies_epures.cend() ; ++it)
					{
						auto identifiant = (*it)->condition_identifiant();
						if(identifiant.first)
						{
							if((*it)->resultat())
							{
								identifiants[identifiant.second].first.push_back(it);
							}
							else
							{
								identifiants[identifiant.second].second.push_back(it);
							}
						}
					}
					std::vector<ConditionSP> operandes_simplifies_epures_sans_doublons;
					for(auto it=operandes_simplifies_epures.cbegin() ; it!=operandes_simplifies_epures.cend() ; ++it)
					{
						auto identifiant = (*it)->condition_identifiant();
						if(identifiant.first)
						{
							if(((not identifiants[identifiant.second].first.empty()) and (it==identifiants[identifiant.second].first.front())) or ((not identifiants[identifiant.second].second.empty()) and (it==identifiants[identifiant.second].second.front())))
							{
								if(not ((not identifiants[identifiant.second].first.empty()) and (not identifiants[identifiant.second].second.empty())))
								{
									operandes_simplifies_epures_sans_doublons.push_back(*it);
								}
							}
						}
						else
						{
							operandes_simplifies_epures_sans_doublons.push_back(*it);
						}
					}
					if(operandes_simplifies_epures_sans_doublons.size()==1)
					{
						return operandes_simplifies_epures_sans_doublons.front();
					}
					ConditionSP condition_simplifiee(new ConditionOuInclusif(operandes_simplifies_epures_sans_doublons));
					return condition_simplifiee;
				}
				virtual bool condition_triviale() const { return false; }
			private:
				explicit ConditionOuInclusif(const std::vector<ConditionSP>& operandes) :_operandes(operandes) {};
				std::vector<ConditionSP> _operandes;
		};

		class ConditionOuExclusif : public Condition
		{
			public:
				ConditionOuExclusif(const ConditionSP& gauche, const ConditionSP& droite)
				{
					_operandes.push_back(gauche);
					_operandes.push_back(droite);
				}
				virtual ~ConditionOuExclusif() { };
				ConditionOuExclusif& operator<<(const ConditionSP& operande)
				{
					_operandes.push_back(operande);
					return *this;
				}
				virtual bool operator() (const Requete& requete)
				{
					bool resultat=false;
					for(auto it=_operandes.cbegin() ; it!=_operandes.cend() ; ++it)
					{
						if((**it)(requete))
						{
							if(resultat)
								return not _resultat;
							else
								resultat = true;
						}
					}
					return (not _resultat) xor resultat;
				}
				virtual ConditionSP simplification() const
				{
					std::vector<ConditionSP> operandes_simplifies;
					for(auto it=_operandes.cbegin() ; it!=_operandes.cend() ; ++it)
					{
						operandes_simplifies.push_back((*it)->simplification());
					}
					std::vector<ConditionSP> operandes_simplifies_epures;
					for(auto it=operandes_simplifies.cbegin() ; it!=operandes_simplifies.cend() ; ++it)
					{
						if((*it)->condition_triviale())
						{
							if((*it)->resultat())
							{
								operandes_simplifies_epures.push_back(*it);
							}
						}
						else
						{
							operandes_simplifies_epures.push_back(*it);
						}
					}
					if(operandes_simplifies_epures.size()==1)
					{
						return operandes_simplifies_epures.front();
					}
					ConditionSP condition_simplifiee(new ConditionOuExclusif(operandes_simplifies_epures));
					return condition_simplifiee;
				}
				virtual bool condition_triviale() const { return false; }
			private:
				explicit ConditionOuExclusif(const std::vector<ConditionSP>& operandes) :_operandes(operandes) {};
				std::vector<ConditionSP> _operandes;
		};
	}
}
#endif
