/*  
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 

#ifndef __INVESTIGATION_DONNEES_HTML__
#define __INVESTIGATION_DONNEES_HTML__

#include <vector>
#include <sstream>
#include <src/global/global.h>
namespace Investigation { namespace Donnees { namespace Html { DECL_SHARED_PTR(Html); } } }
#include <src/global/exceptions.h>
#include <src/global/sources/position.h>

namespace Investigation
{
	namespace Donnees
	{
		namespace Html
		{

			class Attribut
			{
				public:
					Attribut(const std::string& nom, const std::string& valeur)
					:_nom(nom), _valeur(valeur) {};
					friend std::ostream& operator<<(std::ostream& os, const Attribut& attribut)
					{
						os << " " << attribut._nom << "=\"" << attribut._valeur << "\"";
						return os;
					}
				private:
					std::string _nom;
					std::string _valeur;
			};

			typedef std::vector<Attribut> SequenceAttributs;

			DECL_SHARED_PTR(Balise);
			typedef std::vector<BaliseSP> SequenceBalises;
			class Balise
			{
				public:
					virtual ~Balise() {};
					Balise& operator<<(const Attribut& attribut)
					{
						_attributs.push_back(attribut);
						return *this;
					}
					virtual void html(std::ostream& os) const = 0;
				protected:
					Balise(const std::string& nom, const std::string& classe)
					:_nom(nom), _classe(classe) {};
					Balise() {};
					void classe_et_attributs(std::ostream& os) const
					{
						if(not _classe.empty())
						{
							os << " class=\"" << _classe << "\"";
						}
						for(auto it=_attributs.cbegin() ; it!=_attributs.cend() ; ++it)
						{
							os << (*it);
						}
					}
					const std::string _nom;
					const std::string _classe;
					SequenceAttributs _attributs;
			};

		
			class ContenuSimple : public Balise
			{
				public:
					ContenuSimple(const std::string& contenu)
					:_contenu(contenu) {};
					virtual ~ContenuSimple() {};
					virtual void html(std::ostream& os) const
					{
						os << _contenu;
					};
				private:
					const std::string _contenu;
			};

			class BaliseSimple : public Balise
			{
				public:
					explicit BaliseSimple(const std::string& nom, const std::string& classe = "")
					:Balise(nom,classe) {};
					virtual void html(std::ostream& os) const
					{
						os << "<" << _nom;
						classe_et_attributs(os);
						os << "/>";
					};
					using Balise::operator<<;
			};

			class BaliseEncadrante : public Balise
			{
				public:
					explicit BaliseEncadrante(const std::string& nom, const std::string& classe = "")
					:Balise(nom,classe) {};
					virtual ~BaliseEncadrante() {};
					BaliseEncadrante& operator<<(const BaliseEncadrante& baliseencadrante)
					{
						BaliseSP balise(new BaliseEncadrante(baliseencadrante));
						_contenu.push_back(balise);
						return *this;
					}
					BaliseEncadrante& operator<<(const BaliseSimple& balisesimple)
					{
						BaliseSP balise(new BaliseSimple(balisesimple));
						_contenu.push_back(balise);
						return *this;
					}
					BaliseEncadrante& operator<<(const ContenuSimple& contenusimple)
					{
						BaliseSP balise(new ContenuSimple(contenusimple));
						_contenu.push_back(balise);
						return *this;
					}
					BaliseEncadrante& operator<<(const BaliseSP& balise)
					{
						_contenu.push_back(balise);
						return *this;
					}
					using Balise::operator<<;
					virtual void html(std::ostream& os) const
					{
						os << "<" << _nom;
						classe_et_attributs(os);
						os << ">";
						for(auto it=_contenu.cbegin() ; it!=_contenu.cend() ; ++it)
						{
							(*it)->html(os);
						}
						os << "</" << _nom << ">";
					}
				private:
					SequenceBalises _contenu;
			};

			DECL_SHARED_PTR(HtmlElement);

			class HtmlElement
			{
				public:
					virtual ~HtmlElement() {};
					virtual operator BaliseSP () const = 0;
				protected:
					HtmlElement(const std::string& classe)
					:_classe(classe) {};
					std::string _classe;
			};

			typedef std::vector<HtmlElementSP> SequenceHtmlElements;

			class Texte : public HtmlElement
			{
				public:
					Texte(const std::string& texte)
					:HtmlElement(""), _texte(texte) { };
					virtual ~Texte() {};
					virtual operator BaliseSP () const
					{
						BaliseSP balise(new ContenuSimple(_texte));
						return balise;
					};
				private:
					const std::string _texte;
			};

			class A : public HtmlElement
			{
				public:
					explicit A(const std::string& lien, const std::string& classe = "")
					:HtmlElement(classe), _lien(lien) {}
					virtual ~A() {};
					A& operator<<(const Texte& texte)
					{
						HtmlElementSP element(new Texte(texte));
						_elements.push_back(element);
						return *this;
					}
					virtual operator BaliseSP () const
					{
						BaliseEncadrante a("a",_classe);
						a << Attribut("href",_lien);
						for(auto it=_elements.cbegin() ; it!=_elements.cend() ; ++it)
						{
							a << (**it);
						}
						BaliseSP balise(new BaliseEncadrante(a));
						return balise;
					};
				private:
					const std::string _lien;
					SequenceHtmlElements _elements;
			};

			class Hr : public HtmlElement
			{
				public:
					explicit Hr(const std::string& classe = "")
					:HtmlElement(classe) {};
					virtual ~Hr() {};
					virtual operator BaliseSP () const
					{
						BaliseSimple hr("hr",_classe);
						BaliseSP balise(new BaliseSimple(hr));
						return balise;
					};
			};

			class HtmlElementFlux : public HtmlElement
			{
				public:
					explicit HtmlElementFlux(const std::string& classe)
					:HtmlElement(classe) {};
					virtual ~HtmlElementFlux() {};
					HtmlElementFlux& operator<<(const Texte& texte)
					{
						HtmlElementSP element(new Texte(texte));
						_elements.push_back(element);
						return *this;
					}
					HtmlElementFlux& operator<<(const A& a)
					{
						HtmlElementSP element(new A(a));
						_elements.push_back(element);
						return *this;
					}
					friend BaliseEncadrante& operator<<(BaliseEncadrante& balise, const HtmlElementFlux& element)
					{
						for(auto it=element._elements.cbegin() ; it!=element._elements.cend() ; ++it)
						{
							balise << (**it);
						}
						return balise;
					};
				private:
					SequenceHtmlElements _elements;
			};

			class H : public HtmlElementFlux
			{
				public:
					explicit H(const size_t niveau, const std::string& classe = "")
					:HtmlElementFlux(classe), _niveau(niveau) {};
					virtual ~H() {};
					virtual operator BaliseSP () const
					{
						std::ostringstream nom;
						nom << "h" << _niveau;
						BaliseEncadrante h(nom.str(),_classe);
						h << (*this);
						BaliseSP balise(new BaliseEncadrante(h));
						return balise;
					};
				private:
					const size_t _niveau;
			};

			class P : public HtmlElementFlux
			{
				public:
					explicit P(const std::string& classe = "")
					:HtmlElementFlux(classe) {}
					virtual ~P() {};
					virtual operator BaliseSP () const
					{
						BaliseEncadrante p("p",_classe);
						p << (*this);
						BaliseSP balise(new BaliseEncadrante(p));
						return balise;
					};
			};

			class Li : public HtmlElementFlux
			{
				public:
					explicit Li(const std::string& classe = "")
					:HtmlElementFlux(classe) {}
					virtual ~Li() {};
					virtual operator BaliseSP () const
					{
						BaliseEncadrante li("li",_classe);
						li << (*this);
						BaliseSP balise(new BaliseEncadrante(li));
						return balise;
					};
				private:
			};

			typedef std::vector<Li> SequenceLi;

			class Ol : public HtmlElement
			{
				public:
					explicit Ol(const std::string& classe = "")
					:HtmlElement(classe) {};
					Ol& operator<<(const Li& li)
					{
						HtmlElementSP element(new Li(li));
						_elements.push_back(element);
						return *this;
					}
					Ol& operator<<(const SequenceLi& contenu)
					{
						for(auto it=contenu.cbegin() ; it!=contenu.cend() ; ++it)
						{
							(*this) << (*it);
						}
						return *this;
					}
					virtual operator BaliseSP () const
					{
						BaliseEncadrante ul("ol",_classe);
						for(auto it=_elements.cbegin() ; it!=_elements.cend() ; ++it)
						{
							ul << (**it);
						}
						BaliseSP balise(new BaliseEncadrante(ul));
						return balise;
					};
				private:
					SequenceHtmlElements _elements;
			};

			class Ul : public HtmlElement
			{
				public:
					explicit Ul(const std::string& classe = "")
					:HtmlElement(classe) {};
					Ul& operator<<(const Li& li)
					{
						HtmlElementSP element(new Li(li));
						_elements.push_back(element);
						return *this;
					}
					Ul& operator<<(const SequenceLi& contenu)
					{
						for(auto it=contenu.cbegin() ; it!=contenu.cend() ; ++it)
						{
							(*this) << (*it);
						}
						return *this;
					}
					virtual operator BaliseSP () const
					{
						BaliseEncadrante ul("ul",_classe);
						for(auto it=_elements.cbegin() ; it!=_elements.cend() ; ++it)
						{
							ul << (**it);
						}
						BaliseSP balise(new BaliseEncadrante(ul));
						return balise;
					};
				private:
					SequenceHtmlElements _elements;
			};

			class Div;
			class ContenuDiv
			{
				friend class Div;
				public:
					ContenuDiv(const P& p)
					:_element(new P(p)) {};
					ContenuDiv(const Ol& ol)
					:_element(new Ol(ol)) {};
					ContenuDiv(const Ul& ul)
					:_element(new Ul(ul)) {};
					ContenuDiv(const H& h)
					:_element(new H(h)) {};
					ContenuDiv(const Hr& hr)
					:_element(new Hr(hr)) {};
				private:
					HtmlElementSP _element;
			};

			typedef std::vector<ContenuDiv> SequenceContenuDiv;

			class Div : public HtmlElement
			{
				public:
					explicit Div(const std::string& classe = "")
					:HtmlElement(classe) {};
					Div& operator<<(const Div& div)
					{
						HtmlElementSP element(new Div(div));
						_elements.push_back(element);
						return *this;
					}
					Div& operator<<(const ContenuDiv& contenu)
					{
						_elements.push_back(contenu._element);
						return *this;
					}
					Div& operator<<(const SequenceContenuDiv& contenu)
					{
						for(auto it=contenu.cbegin() ; it!=contenu.cend() ; ++it)
						{
							(*this) << (*it);
						}
						return *this;
					}
					virtual operator BaliseSP () const
					{
						BaliseEncadrante div("div",_classe);
						for(auto it=_elements.cbegin() ; it!=_elements.cend() ; ++it)
						{
							div << (**it);
						}
						BaliseSP balise(new BaliseEncadrante(div));
						return balise;
					};
				private:
					SequenceHtmlElements _elements;
			};

			class Html
			{
				public:
					explicit Html(const std::string& titre, const std::string& style = "")
					:_titre(titre), _style(style) {};
					Html& operator<<(const Div& div)
					{
						HtmlElementSP element(new Div(div));
						_elements.push_back(element);
						return *this;
					}
					Html& operator<<(const Hr& hr)
					{
						HtmlElementSP element(new Hr(hr));
						_elements.push_back(element);
						return *this;
					}
					friend std::ostream& operator<<(std::ostream& os, const Html& source)
					{
						BaliseEncadrante html("html");
						BaliseEncadrante head("head");
						BaliseSimple style("link");
						BaliseEncadrante titre("title");
						titre << source._titre;
						head << titre;
						if(not source._style.empty())
						{
							style << Attribut("rel","stylesheet") << Attribut("type","text/css") << Attribut("href",source._style);
							head << style;
						}
						html << head;
						BaliseEncadrante body("body");
						for(auto it=source._elements.cbegin() ; it!=source._elements.cend() ; ++it)
						{
							body << (**it);
						}
						html << body;
						html.html(os);
						return os;
					};
				private:
					const std::string _titre;
					const std::string _style;
					SequenceHtmlElements _elements;
			};
		}
	}
}
#endif
